#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <Files.h>
#include "Win32.h"

//
// Structures
//
typedef struct
{
	char	path[MAX_PATH];
	char	mask[MAX_PATH];
	short   vRefNum;
	long	dirID;
	short   index;
} FIND_FIRST_INFO;

//
// External prototypes
//
extern char *pc2macpath(const char *path);
extern int filelength (FILE *f);

//
// Local variables
//
static char *tempPath = "./Temp/";
OSErr GetPartialPathSpec (const char *partpath, FSSpec *spec);

//
// Local prototypes
//
static int MatchLong (const char *mask, const char *name);



//
// Implementation
//

DWORD WINAPI GetTempPath (DWORD nBufferLength, LPSTR lpBuffer)
{
	if (strlen (tempPath) > nBufferLength) return 0;
	
	strcpy (lpBuffer, tempPath);
	return strlen (tempPath);
}

BOOL WINAPI CopyFile (LPCSTR szSrc, LPCSTR szDst, BOOL fFailIfExists)
{
	FILE *src, *dst;
	size_t fileSize, found;
	Handle buffer;
	OSErr theErr;
	
	// This routine is real bare-bones, just enough for Hexen 2 to work

	// We don't deal with this flag yet
	if (fFailIfExists) return FALSE;
	
	src = dst = NULL;
	buffer = NULL;
	
	// Open the source file
	src = (FILE *) Sys_fopen (szSrc, "rb");
	if (!src)
	{
		Con_Printf ("Couldn't open src file %s for copy.\n", szSrc);
		return FALSE;
	}
	
	// Determine the length of the source file
	fileSize = filelength (src);
	
	// We allocate in the temp heap because the Q1 engine runs pretty tight, memory-wise
	buffer = TempNewHandle (fileSize, &theErr);
	if (!buffer || (theErr != noErr))
	{
		Con_Printf ("Couldn't allocate copy buffer of size %d for copy.\n", fileSize);
		goto bail;
	}
	// We're going to dereference it, lock it down.
	HLock (buffer);
	
	// Read in the data
	found = fread (*buffer, 1, fileSize, src);
	if (found != fileSize) goto bail;
	
	Sys_fclose (src);
	
	dst = (FILE *) Sys_fopen (szDst, "wb");
	if (!dst)
	{
		Con_Printf ("Couldn't open dest file %s for copy.\n", szDst);
		return FALSE;
	}

	
	found = fwrite (*buffer, 1, fileSize, dst);
	if (found != fileSize) goto bail;
	
	// everything was good, close up
	DisposeHandle (buffer);
	Sys_fclose (dst);
	return TRUE;	

bail:
	if (buffer) DisposeHandle (buffer);
	if (src) Sys_fclose (src);
	if (dst) Sys_fclose (dst);
	return FALSE;	
}

BOOL WINAPI DeleteFile (LPCSTR lpFileName)
{
	char *temp;
	
	temp = pc2macpath (lpFileName);
	remove (temp);
	
	return TRUE;
}

HANDLE WINAPI FindFirstFile (LPCSTR lpFileName, LPWIN32_FIND_DATA lpFindFileData)
{
    HANDLE handle;
    FIND_FIRST_INFO *info;
    WIN32_FIND_DATA * data = (WIN32_FIND_DATA *) lpFindFileData;
	char temp[MAX_PATH];
	char *temp1;
	FSSpec pathDir;
	
	// clear out the Win32 structure
    memset (lpFindFileData, 0, sizeof (WIN32_FIND_DATA));

    if (lpFileName == NULL) return 0;
    if (*lpFileName == 0) return 0;
    
    if (!(handle = malloc (sizeof(FIND_FIRST_INFO)))) goto bad;

    info = (FIND_FIRST_INFO *) handle;
    strcpy (temp, lpFileName);

	// split the filename into a path and file mask

	// start by walking from the end of the string, looking for the last
	// pathname separator
	temp1 = temp + strlen (temp);
	while ((*temp1 != '\\') && (*temp1 != '/') && (temp1 != temp))
	{
		temp1 --;
	}
	
	// Did we find a partial (or full) path before the search mask?
	if (temp1 == temp)
	{
		// no path
		info->path[0] = 0;
	}
	else
	{
		// path found
		*temp1++ = 0;

    	// pc2macpath returns a pointer to a local static, so we copy it
		strcpy (info->path, pc2macpath (temp));
	}
	
	// Copy the mask string as well, if it exists
	strcpy (info->mask, temp1);

	// Get a filespec for the path
	GetPartialPathSpec (info->path, &pathDir);
	info->vRefNum = pathDir.vRefNum;
	info->dirID = pathDir.parID;

    info->index = 1;

    if (!FindNextFile (handle, data))
    {
        FindClose (handle);
        goto bad;
    }
    return handle;

bad:
    return INVALID_HANDLE_VALUE;
}

BOOL WINAPI FindNextFile (HANDLE handle, LPWIN32_FIND_DATA lpFindFileData)
{
    FIND_FIRST_INFO *info;
    WIN32_FIND_DATA *data = (WIN32_FIND_DATA *) lpFindFileData;
	CInfoPBRec pb;
	char foundName[MAX_PATH];
	OSErr theErr;

    if ((handle == INVALID_HANDLE_VALUE) ||
        !(info = (FIND_FIRST_INFO *) handle))
    {
        return FALSE;
    }

	while (1)
	{
		// clear the parameter block
		memset (&pb, 0, sizeof(pb));

		pb.hFileInfo.ioNamePtr = (StringPtr) foundName;
		pb.hFileInfo.ioVRefNum = info->vRefNum;
		pb.hFileInfo.ioFDirIndex = info->index;
		pb.hFileInfo.ioDirID = info->dirID;

		// if we find nothing, break out now
		theErr = PBGetCatInfoSync(&pb);
		if (theErr != noErr) break;
		
		p2cstrcpy ((char *)foundName, foundName);
		
		if (strcmp (foundName, lpFindFileData->cFileName) == 0)
		{
			// Same name as the last find, increment the directory ID and start again
			info->index ++;
			continue;
		}
		
		if (MatchLong (info->mask, foundName))
		{
			// we found a match, break
			theErr = noErr;
			break;
		}
		else
		{
			// no match, look for the next one
			info->index ++;
		}
	}
	
	if (theErr == noErr)
	{
		memset (lpFindFileData, 0, sizeof (WIN32_FIND_DATA));
		
		// we found a file, set up our Win32 struct
		// Hexen 2 only checks the name, so that's all we do for now
		strcpy (lpFindFileData->cFileName, foundName);
		return TRUE;
	}
	else
		return FALSE;
}

BOOL WINAPI FindClose (HANDLE handle)
{
    FIND_FIRST_INFO *info;

    if ((handle == INVALID_HANDLE_VALUE) ||
        !(info = (FIND_FIRST_INFO *) handle))
    {
        return FALSE;
    }

	// dispose of our FIND_FIRST_INFO structure
    free (handle);
    
    return TRUE;
}

static int MatchLong (const char *mask, const char *name)
{
    const char *lastjoker = NULL;
    const char *next_to_retry = NULL;

	if (*mask == 0) return 1;
    if (!strcmp( mask, "*.*" )) return 1;

    while (*name && *mask)
    {
        if (*mask == '*')
        {
            mask++;
            while (*mask == '*') mask++;  /* Skip consecutive '*' */
            lastjoker = mask;
            if (!*mask) return 1; /* end of mask is all '*', so match */

            /* skip to the next match after the joker(s) */
            while (*name && (toupper(*name) != toupper(*mask))) name++;

            if (!*name) break;
            next_to_retry = name;
        }
        else if (*mask != '?')
        {
            int mismatch = 0;
            if (toupper(*mask) != toupper(*name)) mismatch = 1;
            if (!mismatch)
            {
                mask++;
                name++;
                if (*mask == '\0')
                {
                    if (*name == '\0')
                        return 1;
                    if (lastjoker)
                        mask = lastjoker;
                }
            }
            else /* mismatch ! */
            {
                if (lastjoker) /* we had an '*', so we can try unlimitedly */
                {
                    mask = lastjoker;

                    /* this scan sequence was a mismatch, so restart
                     * 1 char after the first char we checked last time */
                    next_to_retry++;
                    name = next_to_retry;
                }
                else
                    return 0; /* bad luck */
            }
        }
        else /* '?' */
        {
            mask++;
            name++;
        }
    }
    while ((*mask == '.') || (*mask == '*'))
        mask++;  /* Ignore trailing '.' or '*' in mask */
    return (!*name && !*mask);
}

//
// GetPartialPathSpec
//
// Given a c string (partpath), return a FSSpec that refers
// to the pathname. For exmaple, pass in ":temp" to get a
// FSSpec for the folder 'temp' that is in the app's folder.
// This routine should resolve all aliases
//
OSErr GetPartialPathSpec (const char *partpath, FSSpec *spec)
{
	Str255 path;
	OSErr err;
	FSSpec base;

	memset (&base, 0, sizeof(base));

	sprintf ((char *)path, "%s", partpath);
	c2pstrcpy (path, (char *)path);
	
	err = FSMakeFSSpec (base.vRefNum, base.parID, path, spec);
	if (err == noErr)
	{
		Boolean isFolder, wasAliased;
		err = ResolveAliasFile (spec, true, &isFolder, &wasAliased);
		
		// Our new filespec is for the lowest-level folder in the partial
		// path. We'll look for a fake file in that subfolder to get the
		// directory id of it. Because we look for a fake file, we'll ignore
		// any errors that result.
		p2cstrcpy ((char *) spec->name, spec->name);
		sprintf ((char *)path, ":%s:fake", spec->name);
		c2pstrcpy (path, (char *)path);
		(void)FSMakeFSSpec (spec->vRefNum, spec->parID, path, spec);
	}
	return err;
}

