/*	Wavefront OBJ file loader
	
	adapted from the GLM library (distributed with GLUT) Written by Nate Robins, 1997.
 	email: ndr@pobox.com www: http://www.pobox.com/~ndr
 	
 	Modifications, 'Mac-ifying' and 'Quake-ifying' by Frank Condello
 	email: pox@planetquake.com	www: http://www.planetquake.com/paroxysm/
 
	Changes from GLM:
	- Support for n-sided polygons (up to 32 verts)
	- Request polygon type at loadtime (tris, tris+quads, n-sided)
	- No 'default' group for properly formated OBJ files
	- No material library loading (material name saved as texture reference)
	- Saves Creator info (assumes first comment contains creator info)
	- Translate model (Z going up) for Quake compatibility
*/

#include "mac_io.h"
#include "cmdlib.h"
#include "mathlib.h"

#include "objlib.h"
#include "Matrix.h"

static unsigned	inputflags;
extern qboolean verbose, superverbose;

//
// Post-loading functions - 
//

//
// objTranslate: Translate each vertex in <model> to the matrix <mat>
//
void objTranslate(OBJmodel_t *model, Matrix4 mat)
{
	Vector3D	point;
	int			i;

	if (!model->vertices)
		Error("Model contains no vertices.");
	
	for (i = 1; i <= model->numvertices; i++)
	{
		point = Matrix_MultiplyVertex(mat, model->vertices[X+i*3],
										   model->vertices[Y+i*3],
										   model->vertices[Z+i*3]);
		model->vertices[X+i*3] = point.x;
		model->vertices[Y+i*3] = point.y;
		model->vertices[Z+i*3] = point.z;
	}
}

//
// objGenerateFacetNormals: Generates facet normals (Always build normals from scratch)
//
void objGenerateFacetNormals (OBJmodel_t *model)
{
	int  i;
	vec3_t u, v;
  
	// free any old normals
	if (model->facetnorms)
		free(model->facetnorms);

	// allocate memory for the new normals
	model->numfacetnorms = model->numpolys;
	model->facetnorms = malloc(sizeof(float) * 3 * (model->numfacetnorms + 1));

	for (i = 0; i < model->numpolys; i++)
	{
		T(i).findex = i+1;

		u[X] = model->vertices[3 * T(i).vindices[1] + X] -
			   model->vertices[3 * T(i).vindices[0] + X];
		u[Y] = model->vertices[3 * T(i).vindices[1] + Y] -
			   model->vertices[3 * T(i).vindices[0] + Y];
		u[Z] = model->vertices[3 * T(i).vindices[1] + Z] -
			   model->vertices[3 * T(i).vindices[0] + Z];

		v[X] = model->vertices[3 * T(i).vindices[2] + X] -
			   model->vertices[3 * T(i).vindices[0] + X];
		v[Y] = model->vertices[3 * T(i).vindices[2] + Y] -
			   model->vertices[3 * T(i).vindices[0] + Y];
		v[Z] = model->vertices[3 * T(i).vindices[2] + Z] -
			   model->vertices[3 * T(i).vindices[0] + Z];

	    CrossProduct(u, v, (vec_t *)&model->facetnorms[3 * (i+1)]);
    	VectorNormalize((vec_t *)&model->facetnorms[3 * (i+1)]);
	}
}

//
// objGenerateVertexNormals: Generates smooth vertex normals (perserve edges with angle > 0).
//
void objGenerateVertexNormals(OBJmodel_t* model, float angle)
{
	node_t		*node;
	node_t		*tail;
	node_t		**members;
	float		*normals;
	int			numnormals, numverts;
	vec3_t		average;
	float   	dot, cos_angle = 0;
	int			i, j, avg;
	
	if (!model->facetnorms)
		Error("Can't generate vertex normals for %s. No facet normals found.\n", model->name);

	// calculate the cosine of the angle (in degrees)
	if (angle)
		cos_angle = cos(DEGREES_TO_RADIANS(angle));
		
	// free any old normals
	if (model->normals)
		free(model->normals);

	// allocate space for new normals
	// pOx - we store the number of potential normals at loadtime to deal with n-sided polys
	model->numnormals = model->maxnormals;
	model->normals = malloc(sizeof(float)*3*(model->numnormals+1));

	// allocate a structure that will hold a linked list of poly indices for each vertex
	members = malloc(sizeof(node_t*) * (model->numvertices + 1));
	for (i = 1; i <= model->numvertices; i++)
		members[i] = NULL;
 	
	// for every poly, create a node for each vertex in it
	for (i = 0; i < model->numpolys; i++)
	{	
		numverts = T(i).numverts;
		
		for (j = 0; j < numverts; j++)
		{
			node = malloc(sizeof(node_t));
			node->index = i;
			node->next  = members[T(i).vindices[j]];
			members[T(i).vindices[j]] = node;
		}
	}
	
	// calculate the average normal for each vertex
	numnormals = 1;
	for (i = 1; i <= model->numvertices; i++)
	{
		// calculate an average normal for this vertex by averaging
       	// the facet normal of every polygon this vertex is in
		node = members[i];
		
		if (!node)
		{
			if (superverbose) // Not a critical error
				printf("WARNING: Vertex without a polygon\n");
			continue;
    	}
    	
		average[0] = 0.0; average[1] = 0.0; average[2] = 0.0;
		avg = 0;
		while (node)
		{
			// only average if the angle is less than (or equal to) the threshold angle
			if (cos_angle) {
				dot = DotProduct((&model->facetnorms[3 * T(node->index).findex]),
							     (&model->facetnorms[3 * T(members[i]->index).findex]));
			}
			
			if (!cos_angle || dot > cos_angle)
			{
				node->averaged = true;
				average[0] += model->facetnorms[3 * T(node->index).findex + 0];
				average[1] += model->facetnorms[3 * T(node->index).findex + 1];
				average[2] += model->facetnorms[3 * T(node->index).findex + 2];
				avg = 1;// averaged at least one normal
			}
			else
				node->averaged = false;

			node = node->next;
		}

	    if (avg)
	    {
			// normalize the averaged normal
			VectorNormalize(average);

			// add the normal to the vertex normals list
			model->normals[3 * numnormals + 0] = average[0];
			model->normals[3 * numnormals + 1] = average[1];
			model->normals[3 * numnormals + 2] = average[2];

			avg = numnormals;
			numnormals++;
		}

		// set the normal of this vertex in each poly it is in
		node = members[i];
		while (node)
		{
			if (node->averaged)
			{
				// if this node was averaged, use the average normal
				
				numverts = T(node->index).numverts;
				for (j = 0; j < numverts; j++)
				{
					if (T(node->index).vindices[j] == i) {
						T(node->index).nindices[j] = avg;
						break;
					}
				}
	      	}
			else
	      	{
	      		// if this node wasn't averaged, use the facet normal
				model->normals[3 * numnormals + 0] = 
					model->facetnorms[3 * T(node->index).findex + 0];
				model->normals[3 * numnormals + 1] = 
					model->facetnorms[3 * T(node->index).findex + 1];
				model->normals[3 * numnormals + 2] = 
					model->facetnorms[3 * T(node->index).findex + 2];
					
	      		numverts = T(node->index).numverts;
				for (j = 0; j < numverts; j++)
				{
					if (T(node->index).vindices[j] == i) {
						T(node->index).nindices[j] = numnormals;
					break;
					}
				}
				
				numnormals++;
			}
			node = node->next;
		}
	}
	  
	model->numnormals = numnormals - 1;

	// free the member information
	for (i = 1; i <= model->numvertices; i++)
	{
		node = members[i];
		while (node)
		{
			tail = node;
			node = node->next;
			free(tail);
		}
	}
	free(members);

	// pack the normals array (may or may not be the maximum)
	normals = model->normals;
	model->normals = malloc(sizeof(float) * 3 * (model->numnormals+1));
	for (i = 1; i <= model->numnormals; i++)
	{
		model->normals[3 * i + 0] = normals[3 * i + 0];
		model->normals[3 * i + 1] = normals[3 * i + 1];
		model->normals[3 * i + 2] = normals[3 * i + 2];
	}
	free(normals);

	if (verbose)
		printf(" %d normals generated\n\n", model->numnormals);
}


//
// Group allocation functions
//
OBJgroup_t* objFindGroup(OBJmodel_t* model, char* name)
{
	OBJgroup_t	*group;

	if(!model)
  		return NULL;
  	
	group = model->groups;
	
	while(group)
	{
		if (!strcmp(name, group->name))
			break;
		group = group->next;
	}

	return group;
}

OBJgroup_t* objAddGroup(OBJmodel_t* model, char* name)
{
	OBJgroup_t	*group;

	group = objFindGroup(model, name);
	
	if (!group)
	{
		group = malloc(sizeof(OBJgroup_t));
		group->name = Q_stralloc(name);
		group->texname = Q_stralloc("default");// May be replaced later
		group->numpolys = 0;
		group->polys = NULL;
		group->next = model->groups;
		model->groups = group;
		model->numgroups++;
	}
	
	return group;
}

//
// objCheckVerts: Make sure the current face is within loading limits
//
void objCheckVerts (unsigned numverts)
{
	if ( numverts > OBJ_MAX_POLYVERTS )
		Error ("Polygons cannot contain more than %i verticies.", OBJ_MAX_POLYVERTS);
	
	else if ( numverts > 4 && !(inputflags & U_OBJ_NSIDED) )
		Error ("Model must contain only triangles and quadrangles");
	
	else if ( numverts == 4 && !(inputflags & U_OBJ_QUADS) )
		Error ("Model must contain only triangles.");
}

//
// objFirstPass: get all the model statistics (vertices, normals, etc)
//
static void objFirstPass(OBJmodel_t	*model, FILE* file) 
{
	unsigned int	numvertices,
					numnormals,
					numtexcoords,
					numpolys;					
					
	OBJgroup_t		*group;
	char			buf[256];
		
	numvertices =
	numnormals =
	numtexcoords =
	numpolys = 0;
	
	while(fscanf(file, "%s", buf) != EOF)
	{
		switch(buf[0])
		{
		case '#':						// comment
			// eat up rest of line
			Q_fgets(buf, sizeof(buf), file);
			break;
		
		case 'v':						// v, vn, vt
			switch(buf[1])
			{
			case '\0':					// vertex
				Q_fgets(buf, sizeof(buf), file);
				numvertices++;
				break;
			
			case 'n':					// normal
				Q_fgets(buf, sizeof(buf), file);
				numnormals++;
				break;
			
			case 't':					// texcoord
				Q_fgets(buf, sizeof(buf), file);
				numtexcoords++;
				break;
			
			default:
				printf("objFirstPass(): Unknown token \"%s\".\n", buf);
				exit(1);
				break;
			}
			break;
		
		case 'm':						// material library (not used)
			Q_fgets(buf, sizeof(buf), file);
			break;
    
		case 'u':						// usematerial (texname)
			Q_fgets(buf, sizeof(buf), file);
			break;
    
		case 'g':						// group
			Q_fgets(buf, sizeof(buf), file);
			sscanf(buf, "%s", buf);
			group = objAddGroup(model, buf);
			break;
    
		case 'f':						// face
			Q_fgets(buf, sizeof(buf), file);
			numpolys++;
			
			// Properly formated OBJ files don't need a default group, but just in case...
			if (!group)
				group = objAddGroup(model, "default");
			
			group->numpolys++;

			break;

		default:
			// eat up rest of line
			Q_fgets(buf, sizeof(buf), file);
			break;
		}
	}

	// announce the model statistics
	if (verbose)
	{
		printf(" Vertices: %i\n", numvertices);
		printf(" Normals: %i\n", numnormals);
		printf(" Texcoords: %i\n", numtexcoords);
		printf(" Polygons: %i\n", numpolys);
		printf(" Groups: %i\n\n", model->numgroups);
	}

	// set the stats in the model structure
	model->numvertices  = numvertices;
	model->numnormals   = numnormals;
	model->numtexcoords = numtexcoords;
	model->numpolys = numpolys;

	// allocate memory for the triangles in each group
	group = model->groups;
	while(group)
	{
		group->polys = malloc(sizeof(unsigned int*) * group->numpolys);
		group->numpolys = 0;
		group = group->next;
	}
}


//
// objSecondPass: second pass at a Wavefront OBJ file that gets all
//
static void objSecondPass(OBJmodel_t* model, FILE* file) 
{
	unsigned int	numvertices,
					numnormals,
					numtexcoords,
					numpolys,
					maxnormals;
					
	float			*vertices;
	float			*normals;
	float			*texcoords;
	OBJgroup_t		*group;
	unsigned int	v, n, t, vertcount;
	char     		buf[256], *ptr;

	// set the pointer shortcuts
	vertices = model->vertices;
	normals	= model->normals;
	texcoords = model->texcoords;
	group = model->groups;
	
	// on the second pass through the file, read all the data into the allocated arrays
	numvertices = numnormals = numtexcoords = 1;
	maxnormals = numpolys = 0;
	
	while(fscanf(file, "%s", buf) != EOF)
	{
		switch(buf[0])
		{
		case '#':							// comment
			// pOx - Assume the first comment line contains creator info
			Q_fgets(buf, sizeof(buf), file);
 			if (!model->creator)
	      		model->creator = Q_stralloc(buf);
			break;
		
		case 'v':							// v, vn, vt
			switch(buf[1])
			{
			case '\0':						// vertex
				fscanf(file, "%f %f %f", 
				&vertices[3 * numvertices + X], 
				&vertices[3 * numvertices + Y], 
				&vertices[3 * numvertices + Z]);
				numvertices++;
				break;
			
			case 'n':						// normal
				fscanf(file, "%f %f %f", 
				&normals[3 * numnormals + X],
				&normals[3 * numnormals + Y],
				&normals[3 * numnormals + Z]);
				numnormals++;
				break;
			
			case 't':						// texcoord
				fscanf(file, "%f %f", 
				&texcoords[2 * numtexcoords + X],
				&texcoords[2 * numtexcoords + Y]);
				numtexcoords++;
				break;
			}
			break;
    
    	case 'u':							// material
    		// pOx - changed this to simply save out a
    		// texture name for the current group
			Q_fgets(buf, sizeof(buf), file);
      		sscanf(buf, "%s %s", buf, buf);
      		free (group->texname);// free the default name
      		group->texname = Q_stralloc(buf);
			break;
		
		case 'g':							// group
			// eat up rest of line
			Q_fgets(buf, sizeof(buf), file);
			sscanf(buf, "%s", buf);
			group = objFindGroup(model, buf);
			break;
    
    	// pOx - this bit is a total re-write...
		case 'f':							// face
			v = n = t = vertcount = 0;
			Q_fgets(buf, sizeof(buf), file);
			ptr = &buf[0];
#if 0 
			if (superverbose)// Paranoid - MSL fgets is fucked
				printf("BUF: %s\n", buf);
#endif
			// v//n
			if (strstr(buf, "//"))
			{
				if (superverbose)
					printf ("v//n  ");
					
				while(sscanf(ptr, "%d//%d", &v, &n) > 0)
				{
					T(numpolys).vindices[vertcount] = v;
					T(numpolys).nindices[vertcount] = n;
					
					if (superverbose)
						printf ("%i//%i ", v, n);
						
					vertcount++;
					objCheckVerts (vertcount);
					
					ptr+=3; MOVEPTR(v); MOVEPTR(n);
				}
			}
			
			// v/t/n
			else if (sscanf(buf, "%d/%d/%d", &v, &t, &n) == 3)
			{
				if (superverbose)
					printf ("v/t/n ");
					
				while(sscanf(ptr, "%d/%d/%d", &v, &t, &n) > 0)
				{
					T(numpolys).vindices[vertcount] = v;
					T(numpolys).tindices[vertcount] = t;
					T(numpolys).nindices[vertcount] = n;
					
					if (superverbose)
						printf ("%i/%i/%i ", v, t, n);
					
					vertcount++;
					objCheckVerts (vertcount);
					
					ptr+=3; MOVEPTR(v); MOVEPTR(t); MOVEPTR(n);
				}	
			}
			
			// v/t
			else if (sscanf(buf, "%d/%d", &v, &t) == 2)
			{
				if (superverbose)
					printf ("v/t   ");
				
				while(sscanf(ptr, "%d/%d", &v, &t) > 0)
				{
					T(numpolys).vindices[vertcount] = v;
					T(numpolys).tindices[vertcount] = t;
					
					if (superverbose)
						printf ("%i/%i ", v, t);
					
					vertcount++;
					objCheckVerts (vertcount);
						
					ptr+=2; MOVEPTR(v); MOVEPTR(t);
				}
			}
			
			// v
			else
			{
				if (superverbose)
					printf ("v     ");
					
				while(sscanf(ptr, "%d", &v) > 0)
				{
					T(numpolys).vindices[vertcount] = v;
					
					if (superverbose)
						printf ("%i ", v);
					
					vertcount++;
					objCheckVerts (vertcount);
					
					ptr++; MOVEPTR(v);
				}
			}
			
			group->polys[group->numpolys++] = numpolys;
			
			T(numpolys).numverts = vertcount;
			numpolys++;
			
			maxnormals += vertcount;
			
			if (superverbose)
				printf("\nGroup: %s | Poly: %i | Verts: %i\n\n", group->name, group->numpolys, vertcount);
			
			break;

		default:
			// eat up rest of line
			Q_fgets(buf, sizeof(buf), file);
			break;
		}
	}
	
	// pOx - Need to know how much memory to allocate for normals
	model->maxnormals = maxnormals;
	if (superverbose)
		printf(" %d maximum normals\n\n", model->maxnormals);
	
	// announce the memory requirements
	if (verbose)
	{
		printf(" Memory: %d bytes\n\n",
				numvertices  * 3*sizeof(float) +
				numnormals   * 3*sizeof(float) * (numnormals ? 1 : 0) +
				numtexcoords * 3*sizeof(float) * (numtexcoords ? 1 : 0) +
				numpolys * sizeof(OBJpoly_t));
	}
}


//
// Free model memory
//
void FreeOBJModel (OBJmodel_t *model)
{
	OBJgroup_t	*group;
	
	if(!model)
		return;

	if (model->name) free(model->name);
	if (model->creator) free(model->creator);
	if (model->vertices) free(model->vertices);
	if (model->normals) free(model->normals);
	if (model->texcoords) free(model->texcoords);
	if (model->facetnorms) free(model->facetnorms);
	if (model->polys) free(model->polys);
  
	while(model->groups)
	{
		group = model->groups;
		model->groups = model->groups->next;
		
		if (group->texname) free(group->texname);
		if (group->name) free(group->name);
		if (group->polys) free(group->polys);
		if (group) free(group);
	}

	if (model) free(model);
}


//
// Load OBJ model
//
OBJmodel_t* LoadOBJModel (char *filename, unsigned flags, float smooth, vec3_t translate)
{
  	OBJmodel_t	*model;
  	FILE*     	file;
	Matrix4		mat;
	
	inputflags = flags;
	
	if (!inputflags)
		inputflags = OBJ_TRIANGLES;
	
  	// open the file
  	file = Sys_fopen(filename, "r");
  
  	if (!file)
  	{
    	fprintf(stderr,"Could not open file '%s'\n", filename);
    	exit(1);
  	}
  	
	// allocate a new model & initialize the struct
	model = malloc(sizeof(OBJmodel_t));
	memset (model, 0, sizeof(OBJmodel_t));
	model->name = Q_stralloc(filename);

	// make a first pass through the file to get a count of the number
	// of vertices, normals, texcoords & polys
	objFirstPass(model, file);
	
	// allocate memory
	model->vertices = malloc(sizeof(float) * 3 * (model->numvertices + 1));
	model->polys = malloc(sizeof(OBJpoly_t) * model->numpolys);
	
	if (model->numnormals)
		model->normals = malloc(sizeof(float) * 3 * (model->numnormals + 1));
	if (model->numtexcoords)
		model->texcoords = malloc(sizeof(float) * 2 * (model->numtexcoords + 1));

	// rewind to beginning of file and read in the data this pass
	rewind(file);
	
	objSecondPass(model, file);
	
	Sys_fclose(file);

	// pOx - Default puts Z going up
	Matrix_Identity(&mat);
	mat = Matrix_Rotate(mat, DEGREES_TO_RADIANS(translate[0]),
							 DEGREES_TO_RADIANS(translate[1]),
							 DEGREES_TO_RADIANS(translate[2]));
	objTranslate(model, mat);
	
	objGenerateFacetNormals (model);
	
	if (smooth>-1)// Use -1 for 100% faceted model
		objGenerateVertexNormals (model, smooth);
  	
	return model;
}
